@@ -37,16 +37,61 @@
 #include "utf8.h"
 #include "utils.h"
 
+/* Where to place ellipsis. */
+typedef enum
+{
+	EK_LEFT,   /* On the far left. */
+	EK_RIGHT,  /* On the far right. */
+	EK_MIDDLE, /* In the middle. */
+}
+EllipsisKind;
+
 static int transform_ascii_str(const char str[], int (*f)(int), char buf[],
 		size_t buf_len);
 static int transform_wide_str(const char str[], wint_t (*f)(wint_t), char buf[],
 		size_t buf_len);
 TSTATIC void squash_double_commas(char str[]);
 static char * ellipsis(const char str[], size_t max_width, const char ell[],
-		int right);
+		EllipsisKind kind);
 static size_t copy_substr(char dst[], size_t dst_len, const char src[],
 		char terminator);
 
+void
+str_buf_append(str_buf_t *buf, const char str[])
+{
+	str_buf_append_n(buf, str, strlen(str));
+}
+
+void
+str_buf_append_n(str_buf_t *buf, const char str[], size_t str_len)
+{
+	if(str_len == 0 || buf->data == NULL)
+	{
+		return;
+	}
+
+	char *new_data = realloc(buf->data, buf->len + str_len + 1);
+	if(new_data == NULL)
+	{
+		str_buf_fail(buf);
+		return;
+	}
+
+	buf->data = new_data;
+
+	memcpy(buf->data + buf->len, str, str_len);
+	buf->data[buf->len + str_len] = '\0';
+	buf->len += str_len;
+}
+
+void
+str_buf_fail(str_buf_t *buf)
+{
+	free(buf->data);
+	buf->data = NULL;
+	buf->len = 0;
+}
+
 void
 chomp(char str[])
 {
@@ -282,37 +327,38 @@ transform_ascii_str(const char str[], int (*f)(int), char buf[], size_t buf_len)
 }
 
 /* Transforms characters of the string to while they fit in the buffer by
- * calling specified function on them.  Returns zero on success or non-zero if
- * output buffer is too small. */
+ * calling specified function on them.  Returns zero on success or non-zero on
+ * error (output buffer is too small or out of memory) in which case buffer is
+ * filled with full UTF-8 characters from input. */
 static int
 transform_wide_str(const char str[], wint_t (*f)(wint_t), char buf[],
 		size_t buf_len)
 {
-	size_t copied;
-	int error;
-	wchar_t *wstring;
-	wchar_t *p;
-	char *narrow;
-
-	wstring = to_wide(str);
+	wchar_t *wstring = to_wide(str);
 	if(wstring == NULL)
 	{
 		(void)utf8_strcpy(buf, str, buf_len);
 		return 1;
 	}
 
-	p = wstring;
+	wchar_t *p = wstring;
 	while(*p != L'\0')
 	{
 		*p = f(*p);
 		++p;
 	}
 
-	narrow = to_multibyte(wstring);
-	copied = utf8_strcpy(buf, narrow, buf_len);
-	error = copied == 0U || narrow[copied - 1U] != '\0';
-
+	char *narrow = to_multibyte(wstring);
 	free(wstring);
+
+	if(narrow == NULL)
+	{
+		(void)utf8_strcpy(buf, str, buf_len);
+		return 1;
+	}
+
+	size_t copied = utf8_strcpy(buf, narrow, buf_len);
+	int error = (copied == 0U || narrow[copied - 1U] != '\0');
 	free(narrow);
 
 	return error;
@@ -384,6 +430,18 @@ strnoscmp(const char *s, const char *t, size_t n)
 #endif
 }
 
+int
+strsorter(const void *s, const void *t)
+{
+	return strcmp(*(const char **)s, *(const char **)t);
+}
+
+int
+strcasesorter(const void *s, const void *t)
+{
+	return strcasecmp(*(const char **)s, *(const char **)t);
+}
+
 int
 strossorter(const void *s, const void *t)
 {
@@ -553,23 +611,77 @@ stralign(char str[], size_t width, char pad, int left_align)
 	}
 }
 
+void
+get_left_cut_range(const char str[], size_t max_width, size_t *cut_from,
+		size_t *cut_to)
+{
+	size_t width = utf8_strsw(str);
+	const char *suffix = str;
+
+	while(width > max_width)
+	{
+		width -= utf8_chrsw(suffix);
+		suffix += utf8_chrw(suffix);
+	}
+
+	*cut_from = 0;
+	*cut_to = suffix - str;
+}
+
+void
+get_right_cut_range(const char str[], size_t max_width, size_t *cut_from,
+		size_t *cut_to)
+{
+	const size_t prefix = utf8_nstrsnlen(str, max_width);
+
+	*cut_from = prefix;
+	*cut_to = strlen(str);
+}
+
+void
+get_middle_cut_range(const char str[], size_t max_width, size_t *cut_from,
+		size_t *cut_to)
+{
+	size_t width = utf8_strsw(str);
+	const size_t prefix_width = (max_width + 1)/2;
+
+	const size_t prefix = utf8_nstrsnlen(str, prefix_width);
+	const char *suffix = str + prefix;
+
+	while(width > max_width)
+	{
+		width -= utf8_chrsw(suffix);
+		suffix += utf8_chrw(suffix);
+	}
+
+	*cut_from = prefix;
+	*cut_to = suffix - str;
+}
+
 char *
 left_ellipsis(const char str[], size_t max_width, const char ell[])
 {
-	return ellipsis(str, max_width, ell, 0);
+	return ellipsis(str, max_width, ell, EK_LEFT);
 }
 
 char *
 right_ellipsis(const char str[], size_t max_width, const char ell[])
 {
-	return ellipsis(str, max_width, ell, 1);
+	return ellipsis(str, max_width, ell, EK_RIGHT);
+}
+
+char *
+middle_ellipsis(const char str[], size_t max_width, const char ell[])
+{
+	return ellipsis(str, max_width, ell, EK_MIDDLE);
 }
 
 /* Ensures that str is of width (in character positions) less than or equal to
- * max_width and is aligned appropriately putting ellipsis on one of the ends if
- * needed.  Returns newly allocated modified string. */
+ * max_width and is aligned appropriately putting ellipsis on one of the ends or
+ * in the middle if needed.  Returns newly allocated modified string. */
 static char *
-ellipsis(const char str[], size_t max_width, const char ell[], int right)
+ellipsis(const char str[], size_t max_width, const char ell[],
+		EllipsisKind kind)
 {
 	if(max_width == 0U)
 	{
@@ -592,19 +704,21 @@ ellipsis(const char str[], size_t max_width, const char ell[], int right)
 		return format_str("%.*s", prefix, ell);
 	}
 
-	if(right)
+	size_t cut_from = 0;
+	size_t cut_to = 0;
+
+	if(kind == EK_MIDDLE)
 	{
-		const int prefix = utf8_nstrsnlen(str, max_width - ell_width);
-		return format_str("%.*s%s", prefix, str, ell);
+		get_middle_cut_range(str, max_width - ell_width, &cut_from, &cut_to);
+	}
+	else if(kind == EK_RIGHT)
+	{
+		get_right_cut_range(str, max_width - ell_width, &cut_from, &cut_to);
+	} else {
+		get_left_cut_range(str, max_width - ell_width, &cut_from, &cut_to);
 	}
 
-	while(width > max_width - ell_width)
-	{
-		width -= utf8_chrsw(str);
-		str += utf8_chrw(str);
-	}
-
-	return format_str("%s%s", ell, str);
+	return format_str("%.*s%s%s", (int)cut_from, str, ell, str + cut_to);
 }
 
 char *
@@ -714,7 +828,7 @@ escape_chars(const char string[], const char chars[])
 
 	for(i = 0; i < len; i++)
 	{
-		if(string[i] == '\\' || char_is_one_of(chars, string[i]))
+		if(char_is_one_of(chars, string[i]))
 		{
 			*dup++ = '\\';
 		}
@@ -914,6 +1028,19 @@ read_int(const char line[], int *i)
 
 	*i = (l > INT_MAX) ? INT_MAX : ((l < INT_MIN) ? INT_MIN : l);
 
+	/* XXX: return zero if *i != l? */
+	return *line != '\0' && *endptr == '\0';
+}
+
+int
+read_uint(const char line[], unsigned int *ui)
+{
+	char *endptr;
+	const unsigned long ul = strtoul(line, &endptr, 10);
+
+	*ui = (ul > UINT_MAX ? UINT_MAX : ul);
+
+	/* XXX: return zero if *ui != ul? */
 	return *line != '\0' && *endptr == '\0';
 }
 
@@ -1007,7 +1134,8 @@ split_and_get_dc(char str[], char **state)
 			ptr = str + strlen(str);
 		}
 
-		while(isspace(*str) || *str == ',')
+		/* TODO: maybe drop this loop and adjust callers accordingly. */
+		while(isspace(*str))
 		{
 			++str;
 		}
@@ -1019,7 +1147,7 @@ split_and_get_dc(char str[], char **state)
 			break;
 		}
 
-		str = ptr;
+		str = (ptr[0] == '\0' ? NULL : ptr);
 	}
 
 	if(str == NULL)
@@ -1131,6 +1259,22 @@ double_char(const char str[], char c)
 	return doubled;
 }
 
+int
+str_is_ascii(const char str[])
+{
+	/* Tried to optimize this by checking multiple bytes at once after determining
+	 * size, but it wasn't faster.  The result of such an optimization could
+	 * depend on string length though. */
+	while(*str != '\0')
+	{
+		if((*str++ & 0x80) != 0)
+		{
+			return 0;
+		}
+	}
+	return 1;
+}
+
 #ifndef HAVE_STRCASESTR
 
 char *
